from __future__ import annotations

__all__ = ["ConditionEqualityOperator", "EventHandlerEqualityOperator"]

import logging
from typing import Any

from coola import BaseEqualityOperator, BaseEqualityTester, EqualityTester

from minevent.conditions import BaseCondition
from minevent.handlers import BaseEventHandler

logger = logging.getLogger(__name__)


class ConditionEqualityOperator(BaseEqualityOperator[BaseCondition]):
    r"""Implements an equality operator for ``BaseCondition`` objects.

    Example usage:

    .. code-block:: pycon

        >>> from coola import EqualityTester
        >>> from minevent import PeriodicCondition, ConditionEqualityOperator
        >>> comparator = ConditionEqualityOperator()
        >>> comparator.equal(EqualityTester(), PeriodicCondition(freq=3), PeriodicCondition(freq=3))
        True
        >>> comparator.equal(EqualityTester(), PeriodicCondition(freq=3), PeriodicCondition(freq=2))
        False
    """

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> ConditionEqualityOperator:
        return self.__class__()

    def equal(
        self,
        tester: BaseEqualityTester,
        object1: BaseCondition,
        object2: Any,
        show_difference: bool = False,
    ) -> bool:
        if object1 is object2:
            return True
        if not isinstance(object2, BaseCondition):
            if show_difference:
                logger.info(f"object2 is not a `BaseCondition` object: {type(object2)}")
            return False
        object_equal = object1.equal(object2)
        if show_difference and not object_equal:
            logger.info(
                f"`BaseCondition` objects are different\nobject1=\n{object1}\n"
                f"object2=\n{object2}"
            )
        return object_equal


class EventHandlerEqualityOperator(BaseEqualityOperator[BaseEventHandler]):
    r"""Implements an equality operator for ``BaseEventHandler`` objects.

    Example usage:

    .. code-block:: pycon

        >>> from coola import EqualityTester
        >>> from minevent import EventHandler, EventHandlerEqualityOperator
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> comparator = EventHandlerEqualityOperator()
        >>> comparator.equal(
        ...     EqualityTester(), EventHandler(hello_handler), EventHandler(hello_handler)
        ... )
        True
    """

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> EventHandlerEqualityOperator:
        return self.__class__()

    def equal(
        self,
        tester: BaseEqualityTester,
        object1: BaseEventHandler,
        object2: Any,
        show_difference: bool = False,
    ) -> bool:
        if object1 is object2:
            return True
        if not isinstance(object2, BaseEventHandler):
            if show_difference:
                logger.info(f"object2 is not a `BaseEventHandler` object: {type(object2)}")
            return False
        object_equal = object1.equal(object2)
        if show_difference and not object_equal:
            logger.info(
                f"`BaseEventHandler` objects are different\nobject1=\n{object1}\n"
                f"object2=\n{object2}"
            )
        return object_equal


if not EqualityTester.has_operator(BaseCondition):  # pragma: no cover
    EqualityTester.add_operator(BaseCondition, ConditionEqualityOperator())
if not EqualityTester.has_operator(BaseEventHandler):  # pragma: no cover
    EqualityTester.add_operator(BaseEventHandler, EventHandlerEqualityOperator())



from __future__ import annotations

__all__ = [
    "BaseEventHandler",
    "BaseEventHandlerWithArguments",
    "ConditionalEventHandler",
    "EventHandler",
]

from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from typing import Any

from coola import objects_are_equal
from coola.utils import str_indent, str_mapping

from minevent.conditions import BaseCondition


class BaseEventHandler(ABC):
    r"""Defines the base class to implement an event handler.

    A child class has to implement the following methods:

        - ``handle``
        - ``equal``

    Example usage:

    .. code-block:: pycon

        >>> from minevent import EventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = EventHandler(hello_handler)
        >>> handler
        EventHandler(
          (handler): <function hello_handler at 0x...>
          (handler_args): ()
          (handler_kwargs): {}
        )
        >>> handler.handle()
        Hello!
    """

    def __eq__(self, other: Any) -> bool:
        return self.equal(other)

    @abstractmethod
    def equal(self, other: Any) -> bool:
        r"""Compares two event handlers.

        Args:
        ----
            other: Specifies the other object to compare with.

        Returns:
        -------
            bool: ``True`` if the two event handlers are equal,
                otherwise ``False``.

        Example usage:

        .. code-block:: pycon

            >>> from minevent import EventHandler
            >>> def hello_handler() -> None:
            ...     print("Hello!")
            ...
            >>> handler = EventHandler(hello_handler)
            >>> handler.equal(EventHandler(hello_handler))
            True
            >>> handler.equal(EventHandler(print, handler_args=["Hello!"]))
            False
        """

    @abstractmethod
    def handle(self) -> None:
        r"""Handles the event.

        Example usage:

        .. code-block:: pycon

            >>> from minevent import EventHandler
            >>> def hello_handler() -> None:
            ...     print("Hello!")
            ...
            >>> handler = EventHandler(hello_handler)
            >>> handler.handle()
            Hello!
        """


class BaseEventHandlerWithArguments(BaseEventHandler):
    r"""Defines a base class to implement an event handler with
    positional and/or keyword arguments.

    A child class has to implement the ``equal`` method.

    Args:
    ----
        handler (``Callable``): Specifies the handler.
        handler_args (``Sequence`` or ``None``, optional): Specifies
            the positional argument of the handler.
            Default: ``None``
        handler_kwargs (dict or ``None``, optional): Specifies the
            arbitrary keyword arguments of the handler.
            Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from minevent import EventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = EventHandler(hello_handler)
        >>> handler
        EventHandler(
          (handler): <function hello_handler at 0x...>
          (handler_args): ()
          (handler_kwargs): {}
        )
        >>> handler.handle()
        Hello!
        >>> handler = EventHandler(print, handler_args=["Hello!"])
        >>> handler.handle()
        Hello!
    """

    def __init__(
        self,
        handler: Callable,
        handler_args: Sequence | None = None,
        handler_kwargs: dict | None = None,
    ) -> None:
        if not callable(handler):
            raise TypeError(f"handler is not callable: {handler}")
        self._handler = handler
        self._handler_args = tuple(handler_args or ())
        self._handler_kwargs = handler_kwargs or {}

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "handler": self._handler,
                    "handler_args": self._handler_args,
                    "handler_kwargs": self._handler_kwargs,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def handler(self) -> Callable:
        r"""Callable: The handler."""
        return self._handler

    @property
    def handler_args(self) -> tuple:
        r"""``tuple``: Variable length argument list of the handler."""
        return self._handler_args

    @property
    def handler_kwargs(self) -> dict:
        r"""``dict``: Arbitrary keyword arguments of the handler."""
        return self._handler_kwargs

    def handle(self) -> None:
        self._handler(*self._handler_args, **self._handler_kwargs)


class EventHandler(BaseEventHandlerWithArguments):
    r"""Implements a simple event handler.

    Example usage:

    .. code-block:: pycon

        >>> from minevent import EventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = EventHandler(hello_handler)
        >>> handler
        EventHandler(
          (handler): <function hello_handler at 0x...>
          (handler_args): ()
          (handler_kwargs): {}
        )
        >>> handler.handle()
        Hello!
    """

    def equal(self, other: Any) -> bool:
        if not isinstance(other, EventHandler):
            return False
        return (
            objects_are_equal(self.handler, other.handler)
            and objects_are_equal(self.handler_args, other.handler_args)
            and objects_are_equal(self.handler_kwargs, other.handler_kwargs)
        )


class ConditionalEventHandler(BaseEventHandlerWithArguments):
    r"""Implements a conditional event handler.

    The handler is executed only if the condition is ``True``.

    Args:
    ----
        handler (``Callable``): Specifies the handler.
        condition (``Callable``): Specifies the condition for this
            event handler. The condition should be callable without
            arguments.
        handler_args (``Sequence`` or ``None``, optional): Specifies
            the  positional arguments of the handler.
            Default: ``None``
        handler_kwargs (dict, optional): Specifies the arbitrary
            keyword arguments of the handler. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from minevent import ConditionalEventHandler, PeriodicCondition
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = ConditionalEventHandler(hello_handler, PeriodicCondition(freq=3))
        >>> handler
        ConditionalEventHandler(
          (handler): <function hello_handler at 0x...>
          (handler_args): ()
          (handler_kwargs): {}
          (condition): PeriodicCondition(freq=3, step=0)
        )
        >>> handler.handle()
        Hello!
        >>> handler.handle()
        >>> handler.handle()
        >>> handler.handle()
        Hello!
    """

    def __init__(
        self,
        handler: Callable,
        condition: BaseCondition,
        handler_args: Sequence | None = None,
        handler_kwargs: dict | None = None,
    ) -> None:
        super().__init__(handler=handler, handler_args=handler_args, handler_kwargs=handler_kwargs)
        self._condition = condition

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "handler": self._handler,
                    "handler_args": self._handler_args,
                    "handler_kwargs": self._handler_kwargs,
                    "condition": self._condition,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def condition(self) -> BaseCondition:
        r"""``BaseCondition``: The condition."""
        return self._condition

    def equal(self, other: Any) -> bool:
        if not isinstance(other, ConditionalEventHandler):
            return False
        return (
            objects_are_equal(self.handler, other.handler)
            and objects_are_equal(self.handler_args, other.handler_args)
            and objects_are_equal(self.handler_kwargs, other.handler_kwargs)
            and objects_are_equal(self.condition, other.condition)
        )

    def handle(self) -> None:
        if self._condition.evaluate():
            super().handle()



r"""This module implements an event manager."""

from __future__ import annotations

__all__ = ["EventManager"]

import logging
from collections import defaultdict

from coola.utils import str_indent, str_mapping, str_sequence

from minevent.handlers import BaseEventHandler

logger = logging.getLogger(__name__)


class EventManager:
    r"""Implements an event manager.

    This event manager allows adding event handlers and firing events.
    An event is represented by a case-sensitive string.

    Example usage:

    .. code-block:: pycon

        >>> from minevent import EventHandler, EventManager
        >>> def hello_handler():
        ...     print("Hello!")
        ...
        >>> manager = EventManager()
        >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
        >>> manager.trigger_event("my_event")
        Hello!
    """

    def __init__(self) -> None:
        # This variable is used to store the handlers associated to each event.
        self._event_handlers = defaultdict(list)
        # This variable is used to track the last fired event name
        self._last_triggered_event = None
        self.reset()

    def __repr__(self) -> str:
        event_handlers = str_mapping(
            {
                event: "\n" + str_sequence(handler) if handler else ""
                for event, handler in self._event_handlers.items()
            }
        )
        args = str_indent(
            str_mapping(
                {
                    "event_handlers": "\n" + event_handlers if event_handlers else event_handlers,
                    "last_triggered_event": self._last_triggered_event,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def last_triggered_event(self) -> str | None:
        r"""Gets the last event name that was triggered.

        Returns
        -------
            str or ``None``: The last event name that was fired or
                ``None`` if no event was fired.

        Example usage:

        .. code-block:: pycon

            >>> from minevent import EventHandler, EventManager
            >>> manager = EventManager()
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
            >>> manager.trigger_event("my_event")
            Hello!
            >>> manager.last_triggered_event
        """
        return self._last_triggered_event

    def add_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Adds an event handler to an event.

        The event handler will be called everytime the event happens.

        Args:
        ----
            event (str): Specifies the event to attach the event
                handler.
            event_handler (``BaseEventHandler``): Specifies the
                event handler to attach to the event.

        Example usage:

        .. code-block:: pycon

            >>> from minevent import EventManager, EventHandler
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            >>> manager = EventManager()
            >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
        """
        self._event_handlers[str(event)].append(event_handler)
        logger.debug(f"Added {event_handler} to event {event}")

    def trigger_event(self, event: str) -> None:
        r"""Triggers the handler(s) for the given event.

        Args:
        ----
            event (str): Specifies the event to fire.

        Example usage:

        .. code-block:: pycon

            >>> from minevent import EventHandler, EventManager
            >>> manager = EventManager()
            >>> manager.trigger_event("my_event")  # do nothing because there is no event handler
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
            >>> manager.trigger_event("my_event")
            Hello!
        """
        logger.debug(f"Firing {event} event")
        self._last_triggered_event = event
        for event_handler in self._event_handlers[event]:
            event_handler.handle()

    def has_event_handler(self, event_handler: BaseEventHandler, event: str | None = None) -> bool:
        r"""Indicates if a handler is registered in the event manager.

        Note that this method relies on the ``equal`` method of the
        input event handler to compare event handlers.

        Args:
        ----
            event_handler (``BaseEventHandler``): Specifies the eventn
                handler to check.
            event (str or ``None``): Specifies an event to check. If
                the value is ``None``, it will check all the events.
                Default: ``None``

        Example usage:

        .. code-block:: pycon

            >>> from minevent import EventHandler, EventManager
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            >>> manager = EventManager()
            >>> # Check if `hello_handler` is registered in the event manager
            >>> manager.has_event_handler(EventHandler(hello_handler))
            False
            >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
            False
            >>> # Add an event handler
            >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
            >>> # Check if `hello_handler` is registered in the event manager
            >>> manager.has_event_handler(EventHandler(hello_handler))
            True
            >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
            True
            >>> # Check if `hello_handler` is registered in the event manager for 'my_other_event' event
            >>> manager.has_event_handler(EventHandler(hello_handler), "my_other_event")
            False
        """
        events = [event] if event else self._event_handlers
        for evnt in events:
            for handler in self._event_handlers[evnt]:
                if event_handler.equal(handler):
                    return True
        return False

    def remove_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Removes an event handler of a given event.

        Note that if the same event handler was added multiple times
        the event, all the duplicated handlers are removed. This
        method relies on the ``equal`` method of the input event
        handler to compare event handlers.

        Args:
        ----
            event (str): Specifies the event handler is attached to.
            event_handler (``BaseEventHandler``): Specifies the event
                handler to remove.

        Raises:
        ------
            ValueError: if the event does not exist or if the handler
                is not attached to the event.

        Example usage:

        .. code-block:: pycon

            >>> from minevent import EventHandler, EventManager
            >>> manager = EventManager()
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
            >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
            True
            >>> # Remove the event handler of the engine
            >>> manager.remove_event_handler("my_event", EventHandler(hello_handler))
            >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
            False
        """
        if event not in self._event_handlers:
            raise RuntimeError(f"'{event}' event does not exist")

        new_event_handlers = [
            handler for handler in self._event_handlers[event] if not event_handler.equal(handler)
        ]
        if len(new_event_handlers) == len(self._event_handlers[event]):
            raise RuntimeError(
                f"{event_handler} is not found among registered event handlers for '{event}' event"
            )
        if len(new_event_handlers) > 0:
            self._event_handlers[event] = new_event_handlers
        else:
            del self._event_handlers[event]
        logger.debug(f"Removed {event_handler} in '{event}' event")

    def reset(self) -> None:
        r"""Resets the event manager.

        This method removes all the event handlers from the event manager.

        Example usage:

        .. code-block:: pycon

            >>> # Create an event manager
            >>> from minevent import EventManager
            >>> manager = EventManager()
            >>> # Add an event handler to the engine
            >>> def hello_handler():
            ...     print("Hello!")
            ...
            >>> from minevent import EventHandler
            >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
            >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
            True
            >>> manager.trigger_event("my_event")
            >>> manager.last_triggered_event
            my_event
            >>> # Reset the event manager
            >>> manager.reset()
            >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
            >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
            False
            >>> manager.last_triggered_event
            None
        """
        self._event_handlers.clear()
        self._last_triggered_event = None
